# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


class OutOfDisk(Exception):
    pass


def get_app_dir():
    import os
    conda_prefix = os.environ.get('CONDA_PREFIX')
    if conda_prefix is not None and os.access(conda_prefix, os.W_OK | os.X_OK):
        return os.path.join(conda_prefix, 'var', 'q2cli')
    else:
        import click
        return click.get_app_dir('q2cli', roaming=False)


# NOTE: `get_cache_dir` and `get_completion_path` live here instead of
# `q2cli.cache` because `q2cli.cache` can be  slow to import.
# `get_completion_path` (which relies on `get_cache_dir`) is imported and
# executed by the Bash completion function each time the user hits <tab>, so it
# must be quick to import.
def get_cache_dir():
    import os.path
    return os.path.join(get_app_dir(), 'cache')


def get_completion_path():
    import os.path
    return os.path.join(get_cache_dir(), 'completion.sh')


def hidden_to_cli_name(name):
    # Safety first
    if not name.startswith('_'):
        raise ValueError(f"The name '{name}' does not start with '_' meaning"
                         " it is not a hidden action and this method should"
                         " not have been called on it.")

    name = to_cli_name(name)

    # Retain the leading _
    return name.replace('-', '_', 1)


def to_cli_name(name):
    return name.replace('_', '-')


def to_snake_case(name):
    return name.replace('-', '_')


def exit_with_error(e, header='An error has been encountered:',
                    traceback='stderr', status=1):
    import sys
    import traceback as tb
    import textwrap
    import click
    from q2cli.core.config import CONFIG

    footer = []  # footer only exists if traceback is set
    tb_file = None
    if traceback == 'stderr':
        tb_file = sys.stderr
        footer = ['See above for debug info.']
    elif traceback is not None:
        tb_file = traceback
        footer = ['Debug info has been saved to %s' % tb_file.name]

    error = textwrap.indent(str(e), '  ')
    segments = [header, error] + footer

    if traceback is not None:
        tb.print_exception(type(e), e, e.__traceback__, file=tb_file)

        tb_file.write('\n')

    click.echo(CONFIG.cfg_style('error', '\n\n'.join(segments)), err=True)

    if not footer:
        click.echo(err=True)  # extra newline to look normal

    try:
        click.get_current_context().exit(status)
    except RuntimeError:
        sys.exit(status)


def output_in_cache(fp):
    """Determines if an output path follows the format
    /path_to_extant_cache:key
    """
    from qiime2.core.cache import Cache

    # Tells us right away this isn't in a cache
    if ':' not in fp:
        return False

    cache_path, key = _get_cache_path_and_key(fp)

    try:
        if Cache.is_cache(cache_path):
            Cache.validate_key(key)
            return True
    except FileNotFoundError as e:
        # If cache_path doesn't exist, don't treat this as a cache output
        if 'No such file or directory' in str(e):
            pass
        else:
            raise e

    # We don't have a cache at all
    return False


def get_close_matches(name, possibilities):
    import difflib

    name = name.lower()
    # bash completion makes an incomplete arg most likely
    matches = [m for m in possibilities if m.startswith(name)]
    if not matches:
        # otherwise, it may be misspelled
        matches = difflib.get_close_matches(name, possibilities, cutoff=0.8)

    matches.sort()

    if len(matches) > 5:
        # this is probably a good time to look at --help
        matches = matches[:4] + ['...']

    return matches


class pretty_failure:
    def __init__(self, header='An error has been encountered:',
                 traceback='stderr', status=1):
        self.header = header
        self.traceback = traceback
        self.status = status

    def __call__(self, function):
        def wrapped(*args, **kwargs):
            with self:
                return function(*args, **kwargs, failure=self)
        # not using functools.wraps to keep import overhead low
        # click only seems to need the __name__
        wrapped.__name__ = function.__name__
        return wrapped

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if exit_with_error is called twice, then click.exit(1) or sys.exit(1)
        # will happen, no need to exit_with_error again in that case.
        if exc_val is not None and str(exc_val) != '1':
            exit_with_error(exc_val, self.header, self.traceback, self.status)

        return False


def convert_primitive(ast):
    import click

    mapping = {
        'Int': int,
        'Str': str,
        'Float': float,
        'Color': str,
        'Bool': bool
    }
    # TODO: it would be a good idea to refactor this someday, but until then
    # just handle the few predicates we know about.
    predicate = ast['predicate']
    if predicate:
        if predicate['name'] == 'Choices' and ast['name'] == 'Str':
            return click.Choice(predicate['choices'])
        elif predicate['name'] == 'Range' and ast['name'] == 'Int':
            start = predicate['range'][0]
            end = predicate['range'][1]
            # click.IntRange is always inclusive
            if start is not None and not predicate['inclusive'][0]:
                start += 1
            if end is not None and not predicate['inclusive'][1]:
                end -= 1
            return click.IntRange(start, end)
        elif predicate['name'] == 'Range' and ast['name'] == 'Float':
            # click.FloatRange will be in click 7.0, so for now the
            # range handling will just fallback to qiime2.
            return mapping['Float']
        else:
            raise NotImplementedError()
    else:
        return mapping[ast['name']]


def citations_option(get_citation_records):
    import click
    from q2cli.core.config import CONFIG

    def callback(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return

        records = get_citation_records()
        if records:
            import io
            import qiime2.sdk

            citations = qiime2.sdk.Citations(
                [('key%d' % i, r) for i, r in enumerate(records)])
            with io.StringIO() as fh:
                fh.write('% use `qiime tools citations` on a QIIME 2 result'
                         ' for complete list\n\n')
                citations.save(fh)
                click.echo(fh.getvalue(), nl=False)
            ctx.exit()
        else:
            click.echo(
                CONFIG.cfg_style('problem', 'No citations found.'), err=True)
            ctx.exit(1)

    return click.Option(['--citations'], is_flag=True, expose_value=False,
                        is_eager=True, callback=callback,
                        help='Show citations and exit.')


def example_data_option(get_plugin, action_name=None):
    import click
    from q2cli.click.type import OutDirType

    def callback(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        else:
            import q2cli.core.usage as usage

        plugin = get_plugin()
        if action_name is not None:
            action = plugin.actions[action_name]
            generator = usage.write_example_data(action, value)
        else:
            generator = usage.write_plugin_example_data(plugin, value)

        ran = False
        for hint, path in generator:
            click.secho('Saved %s to: %s' % (hint, path), fg='green')
            ran = True

        if ran:
            ctx.exit()
        else:
            click.secho('No example data found.', fg='yellow', err=True)
            ctx.exit(1)

    return click.Option(['--example-data'], type=OutDirType(), is_eager=True,
                        expose_value=False, callback=callback,
                        help='Write example data and exit.')


def get_plugin_manager():
    import qiime2.sdk

    try:
        return qiime2.sdk.PluginManager.reuse_existing()
    except qiime2.sdk.UninitializedPluginManagerError:
        import os

        if 'MYSTERY_STEW' in os.environ:
            from q2_mystery_stew.plugin_setup import create_plugin

            the_stew = create_plugin()
            pm = qiime2.sdk.PluginManager(add_plugins=False)
            pm.add_plugin(the_stew)
            return pm

        return qiime2.sdk.PluginManager()


def load_metadata(fp):
    import qiime2
    import sys

    metadata, error = _load_metadata_artifact(fp)
    if metadata is None:
        try:
            metadata = qiime2.Metadata.load(fp)
        except Exception as e:
            if error and ':' in fp:
                e = error
            header = ("There was an issue with loading the file %s as "
                      "metadata:" % fp)
            tb = 'stderr' if '--verbose' in sys.argv else None
            exit_with_error(e, header=header, traceback=tb)

    return metadata


def _load_metadata_artifact(fp):
    import qiime2
    import sys

    artifact, error = _load_input(fp)
    artifact = artifact[1]
    if isinstance(error, OutOfDisk):
        raise error

    default_tb = 'stderr'
    # if that worked, we have an artifact or we've
    # already raised a critical error
    # otherwise, any normal errors can be ignored as its
    # most likely actually metadata not a qza
    if artifact:
        try:
            default_tb = None
            if isinstance(artifact, qiime2.Visualization):
                raise Exception(
                    'Visualizations cannot be viewed as QIIME 2 metadata.')
            if not artifact.has_metadata():
                raise Exception(
                    f"Artifacts with type {artifact.type!r} cannot be viewed"
                    " as QIIME 2 metadata.")

            default_tb = 'stderr'
            return artifact.view(qiime2.Metadata), None

        except Exception as e:
            header = ("There was an issue with viewing the artifact "
                      f"{fp!r} as QIIME 2 Metadata:")
            tb = 'stderr' if '--verbose' in sys.argv else default_tb
            exit_with_error(e, header=header, traceback=tb)

    else:
        return None, error


def _load_input(fp, view=False):
    # Just initialize the plugin manager. This is slow and not necessary if we
    # called this from qiime tools view.
    import os

    key = None

    if not view:
        _ = get_plugin_manager()

    # We are loading a collection from outside of a cache. This cannot be keyed
    if os.path.isdir(fp):
        if len(os.listdir(fp)) == 0:
            raise ValueError(f"Provided directory '{fp}' is empty.")

        artifact, error = _load_collection(fp)
    # We may be loading something from a cache with or without and additional
    # key, or we may be loading a piece of data from outside of a cache with a
    # key. We could also be loading a normal unkeyed artifact with a : in its
    # path
    elif ':' in fp:
        # First we assume this is just a weird filepath
        artifact, _ = _load_input_file(fp)
        # Then we check if it is a key:path
        if artifact is None:
            key, new_fp = _get_path_and_collection_key(fp)
            artifact, _ = _load_input_file(new_fp)

        # If we still have nothing
        if artifact is None:
            key = None
            # We assume this is a cache:key. We keep this error because we
            # assume if they had a : in their path they were trying to load
            # something from a cache
            artifact, error = _load_input_cache(fp)
            if error:
                # Then we check if it is a key:cache:key
                key, new_fp = _get_path_and_collection_key(fp)
                artifact, _ = _load_input_cache(new_fp)

        # If we ended up with an artifact, we disregard our error
        if artifact is not None:
            error = None
    # We are just loading a normal artifact on disk without silly colons in the
    # filepath
    else:
        artifact, error = _load_input_file(fp)

    if isinstance(error, OSError) and error.errno == 28:
        # abort as there's nothing anyone can do about this
        from qiime2.core.cache import get_cache

        path = str(get_cache().path)
        return None, OutOfDisk(f'There was not enough space left on {path!r} '
                               f'to use the artifact {fp!r}. (Try '
                               f'setting $TMPDIR to a directory with more '
                               f'space, or increasing the size of {path!r})')

    return (key, artifact), error


# NOTE: These load collection functions are now virtually identical to class
# methods on qiime2.sdk.Result
def _load_collection(fp):
    import os
    import warnings

    order_fp = os.path.join(fp, '.order')

    if os.path.isfile(order_fp):
        artifacts, error = _load_ordered_collection(fp, order_fp)
    else:
        warnings.warn(f'The directory {fp} does not contain a .order file. '
                      'The files will be read into the collection in the '
                      'order the filesystem provides them in.')
        artifacts, error = _load_unordered_collection(fp)

    return artifacts, error


def _load_ordered_collection(fp, order_fp):
    import os

    artifacts = {}

    with open(order_fp) as order_fh:
        for key in order_fh.read().splitlines():
            artifact_path = os.path.join(fp, f'{key}.qza')
            artifacts[key], error = _load_input_file(artifact_path)

            if error:
                return None, error

    return artifacts, None


def _load_unordered_collection(fp):
    import os

    artifacts = {}

    for artifact in os.listdir(fp):
        artifact_fp = os.path.join(fp, artifact)
        artifacts[artifact], error = _load_input_file(artifact_fp)

        if error:
            return None, error

    return artifacts, None


def _load_input_cache(fp):
    artifact = error = None
    try:
        artifact = try_as_cache_input(fp)
    except Exception as e:
        error = e

    return artifact, error


def _load_input_file(fp):
    import qiime2.sdk
    from os.path import expanduser

    # If there is a leading ~ we expand it to be the path to home
    fp = expanduser(fp)

    # test if valid
    peek = None
    try:
        peek = qiime2.sdk.Result.peek(fp)
    except Exception as error:
        if isinstance(error, SyntaxError):
            raise error
        # ideally ValueError: X is not a QIIME archive.
        # but sometimes SyntaxError or worse
        return None, error

    # try to actually load
    try:
        artifact = qiime2.sdk.Result.load(fp)
        return artifact, None

    except Exception as e:
        if peek:
            # abort early as there's nothing else to do
            raise ValueError(
                "It looks like you have an Artifact but are missing the"
                " plugin(s) necessary to load it. Artifact has type"
                f" {peek.type!r} and format {peek.format!r}") from e
        else:
            error = e

        return None, error


def try_as_cache_input(fp):
    """ Determine if an input is in a cache and load it from the cache if it is
    """
    import os
    from qiime2 import Cache

    cache_path, key = _get_cache_path_and_key(fp)

    # We don't want to invent a new cache on disk here because if their input
    # exists their cache must also already exist
    if not os.path.exists(cache_path) or not Cache.is_cache(cache_path):
        raise ValueError(f"The path {cache_path!r} is not a valid cache.")

    cache = Cache(cache_path)
    return cache.load(key)


def _get_cache_path_and_key(fp):
    return fp.rsplit(':', 1)


def _get_path_and_collection_key(fp):
    return fp.split(':', 1)


def get_default_recycle_pool(plugin_action):
    from hashlib import sha1

    return f'recycle_{plugin_action}_' \
           f'{sha1(plugin_action.encode("utf-8")).hexdigest()}'



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
from typing import Literal

import click

import q2cli.util
from q2cli.click.command import ToolCommand, ToolGroupCommand

_COMBO_METAVAR = 'ARTIFACT/VISUALIZATION'


@click.group(help='Tools for working with QIIME 2 files.',
             cls=ToolGroupCommand)
def tools():
    pass


@tools.command(name='export',
               short_help='Export data from a QIIME 2 Artifact '
               'or a Visualization',
               help='Exporting extracts (and optionally transforms) data '
               'stored inside an Artifact or Visualization. Note that '
               'Visualizations cannot be transformed with --output-format',
               cls=ToolCommand)
@click.option('--input-path', required=True, metavar=_COMBO_METAVAR,
              type=click.Path(exists=True, file_okay=True,
                              dir_okay=False, readable=True),
              help='Path to file that should be exported')
@click.option('--output-path', required=True,
              type=click.Path(exists=False, file_okay=True, dir_okay=True,
                              writable=True),
              help='Path to file or directory where '
              'data should be exported to')
@click.option('--output-format', required=False,
              help='Format which the data should be exported as. '
              'This option cannot be used with Visualizations')
def export_data(input_path, output_path, output_format):
    import qiime2.util
    import qiime2.sdk
    import distutils
    from q2cli.core.config import CONFIG
    result = qiime2.sdk.Result.load(input_path)
    if output_format is None:
        if isinstance(result, qiime2.sdk.Artifact):
            output_format = result.format.__name__
        else:
            output_format = 'Visualization'
        result.export_data(output_path)
    else:
        if isinstance(result, qiime2.sdk.Visualization):
            error = '--output-format cannot be used with visualizations'
            click.echo(CONFIG.cfg_style('error', error), err=True)
            click.get_current_context().exit(1)
        else:
            source = result.view(qiime2.sdk.parse_format(output_format))
            if os.path.isfile(str(source)):
                if os.path.isfile(output_path):
                    os.remove(output_path)
                elif os.path.dirname(output_path) == '':
                    # This allows the user to pass a filename as a path if they
                    # want their output in the current working directory
                    output_path = os.path.join('.', output_path)
                if os.path.dirname(output_path) != '':
                    # create directory (recursively) if it doesn't exist yet
                    os.makedirs(os.path.dirname(output_path), exist_ok=True)
                qiime2.util.duplicate(str(source), output_path)
            else:
                distutils.dir_util.copy_tree(str(source), output_path)

    output_type = 'file' if os.path.isfile(output_path) else 'directory'
    success = 'Exported %s as %s to %s %s' % (input_path, output_format,
                                              output_type, output_path)
    click.echo(CONFIG.cfg_style('success', success))


def _print_descriptions(descriptions, tsv):
    if tsv:
        for value, description in descriptions.items():
            click.echo(f"{value}\t", nl=False)
            if description:
                click.echo(_deformat_description(description))
            else:
                click.echo()
    else:
        import textwrap
        tabsize = 8
        for value, description in descriptions.items():
            click.secho(value, bold=True)
            if description:
                description = _deformat_description(description)
                wrapped_description = textwrap.wrap(description,
                                                    width=72-tabsize,
                                                    initial_indent='\t',
                                                    subsequent_indent='\t',
                                                    tabsize=tabsize)
                for line in wrapped_description:
                    click.echo(f"{line}")
            else:
                click.secho("\tNo description", italic=True)
            click.echo()


def _deformat_description(description):
    import re
    deformatted = re.sub(r"[\t\n]+", ' ', description)
    despaced = re.sub(r" +", ' ', deformatted)
    return despaced


def _get_matches(words, possibilities, strict=False):
    from difflib import get_close_matches
    if strict:
        cutoff = 1
    else:
        cutoff = 0.6

    matches = set()
    num_possibilities = len(possibilities)
    for word in words:
        matches.update(get_close_matches(word,
                                         possibilities,
                                         n=num_possibilities,
                                         cutoff=cutoff))
        # substring search
        if cutoff != 1:
            for possibility in possibilities:
                if word.lower() in possibility.lower():
                    matches.add(possibility)

    return list(matches)


@tools.command(
        name='list-types',
        help='List the available semantic types.',
        short_help='',
        cls=ToolCommand
)
@click.argument('queries', nargs=-1)
@click.option('--strict', is_flag=True,
              help='Show only exact matches for the type argument(s).')
@click.option('--tsv', is_flag=True,
              help='Print as machine readable tab-separated values.')
def show_types(queries, strict, tsv):
    pm = q2cli.util.get_plugin_manager()

    if len(queries) > 0:
        matches = _get_matches(queries, list(pm.artifact_classes), strict)
    else:
        matches = sorted(list(pm.artifact_classes))

    descriptions = {}
    for match in matches:
        description = pm.artifact_classes[match].description
        descriptions[match] = description

    _print_descriptions(descriptions, tsv)


@tools.command(
        name='list-formats',
        help='List the available formats.',
        short_help='',
        cls=ToolCommand
)
@click.argument('queries', nargs=-1)
@click.option('--importable', is_flag=True,
              help='List the importable formats.')
@click.option('--exportable', is_flag=True,
              help='List the exportable formats.')
@click.option('--strict', is_flag=True,
              help='Show only exact matches for the format argument(s).')
@click.option('--tsv', is_flag=True,
              help='Print as machine readable tab-separated values.')
def show_formats(queries, importable, exportable, strict, tsv):
    if importable and exportable:
        raise click.UsageError("'--importable' and '--exportable' flags are "
                               "mutually exclusive.")
    if not importable and not exportable:
        raise click.UsageError("One of '--importable' or '--exportable' flags "
                               "is required.")

    pm = q2cli.util.get_plugin_manager()
    portable_formats = pm.importable_formats if importable \
        else pm.exportable_formats

    if len(queries) > 0:
        matches = _get_matches(queries, portable_formats.keys(), strict)
    else:
        matches = sorted(portable_formats.keys())

    descriptions = {}
    for match in matches:
        docstring = portable_formats[match].format.__doc__
        first_docstring_line = docstring.split('\n\n')[0].strip() \
            if docstring else ''
        descriptions[match] = first_docstring_line

    _print_descriptions(descriptions, tsv)


@tools.command(name='import',
               short_help='Import data into a new QIIME 2 Artifact.',
               help="Import data to create a new QIIME 2 Artifact. See "
                    "https://docs.qiime2.org/ for usage examples and details "
                    "on the file types and associated semantic types that can "
                    "be imported.",
                    cls=ToolCommand)
@click.option('--type', required=True,
              help='The semantic type of the artifact that will be created '
                   'upon importing. Use `qiime tools list-types` to see what '
                   'importable semantic types are available in the current '
                   'deployment.')
@click.option('--input-path', required=True,
              type=click.Path(exists=True, file_okay=True, dir_okay=True,
                              readable=True),
              help='Path to file or directory that should be imported.')
@click.option('--output-path', required=True, metavar='ARTIFACT',
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              writable=True),
              help='Path where output artifact should be written.')
@click.option('--input-format', required=False,
              help='The format of the data to be imported. If not provided, '
                   'data must be in the format expected by the semantic type '
                   'provided via --type. Use `qiime tools list-formats '
                   '--importable` to see which formats of input data are '
                   'importable.')
@click.option('--validate-level', default='max',
              type=click.Choice(['min', 'max']),
              help='How much to validate the imported data before creating the'
                   ' artifact. A value of "max" will generally read the entire'
                   ' file or directory, whereas "min" will not usually do so.'
                   ' [default: "max"]')
def import_data(type, input_path, output_path, input_format, validate_level):

    from q2cli.core.config import CONFIG

    artifact = _import(type, input_path, input_format, validate_level)
    artifact.save(output_path)

    if input_format is None:
        input_format = artifact.format.__name__

    success = 'Imported %s as %s to %s' % (input_path,
                                           input_format,
                                           output_path)
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(short_help='Take a peek at a QIIME 2 Artifact or '
                          'Visualization.',
               help="Display basic information about a QIIME 2 Artifact or "
                    "Visualization, including its UUID and type.",
               cls=ToolCommand)
@click.argument('paths', nargs=-1, required=True,
                type=click.Path(exists=True, file_okay=True, dir_okay=False,
                                readable=True), metavar=_COMBO_METAVAR)
@click.option('--tsv/--no-tsv', default=False,
              help='Print as machine-readable tab-separated values.')
def peek(paths, tsv):
    import qiime2.sdk
    from q2cli.core.config import CONFIG

    metadatas = {os.path.basename(path):
                 qiime2.sdk.Result.peek(path) for path in paths}

    if tsv:
        click.echo("Filename\tType\tUUID\tData Format")
        for path, m in metadatas.items():
            click.echo(f"{path}\t{m.type}\t{m.uuid}\t{m.format}")

    elif len(metadatas) == 1:
        metadata = metadatas[os.path.basename(paths[0])]
        click.echo(CONFIG.cfg_style('type', "UUID")+":        ", nl=False)
        click.echo(metadata.uuid)
        click.echo(CONFIG.cfg_style('type', "Type")+":        ", nl=False)
        click.echo(metadata.type)
        if metadata.format is not None:
            click.echo(CONFIG.cfg_style('type', "Data format")+": ", nl=False)
            click.echo(metadata.format)

    else:
        COLUMN_FILENAME = "Filename"
        COLUMN_TYPE = "Type"
        COLUMN_UUID = "UUID"
        COLUMN_DATA_FORMAT = "Data Format"

        filename_width = max([len(p) for p in paths]
                             + [len(COLUMN_FILENAME)])
        type_width = max([len(i.type) for i in metadatas.values()]
                         + [len(COLUMN_TYPE)])
        uuid_width = max([len(i.uuid) for i in metadatas.values()]
                         + [len(COLUMN_UUID)])
        data_format_width = \
            max([len(i.format) if i.format is not None else 0
                 for i in metadatas.values()] + [len(COLUMN_DATA_FORMAT)])

        padding = 2
        format_string = f"{{f:<{filename_width + padding}}} " + \
                        f"{{t:<{type_width + padding}}} " + \
                        f"{{u:<{uuid_width + padding}}} " + \
                        f"{{d:<{data_format_width + padding}}}"

        click.secho(
            format_string.format(
                f=COLUMN_FILENAME,
                t=COLUMN_TYPE,
                u=COLUMN_UUID,
                d=COLUMN_DATA_FORMAT),
            bold=True, fg="green")
        for path, m in metadatas.items():
            click.echo(
                format_string.format(
                    f=path,
                    t=m.type,
                    u=m.uuid,
                    d=(m.format if m.format is not None else 'N/A')))


_COLUMN_TYPES = ['categorical', 'numeric']


@tools.command(name='cast-metadata',
               short_help='Designate metadata column types.',
               help='Designate metadata column types.'
                    ' Supported column types are as follows: %s.'
                    ' Providing multiple file paths to this command will merge'
                    ' the metadata.' % (', '.join(_COLUMN_TYPES)),
               cls=ToolCommand)
@click.option('--cast', required=True, metavar='COLUMN:TYPE', multiple=True,
              help='Parameter for each metadata column that should'
              ' be cast as a specified column type (supported types are as'
              ' follows: %s). The required formatting for this'
              ' parameter is --cast COLUMN:TYPE, repeated for each column'
              ' and the associated column type it should be cast to in'
              ' the output.' % (', '.join(_COLUMN_TYPES)))
@click.option('--ignore-extra', is_flag=True,
              help='If this flag is enabled, cast parameters that do not'
              ' correspond to any of the column names within the provided'
              ' metadata will be ignored.')
@click.option('--error-on-missing', is_flag=True,
              help='If this flag is enabled, failing to include cast'
              ' parameters for all columns in the provided metadata will'
              ' result in an error.')
@click.option('--output-file', required=False,
              type=click.Path(exists=False, file_okay=True, dir_okay=False,
                              writable=True),
              help='Path to file where the modified metadata should be'
              ' written to.')
@click.argument('paths', nargs=-1, required=True, metavar='METADATA...',
                type=click.Path(exists=True, file_okay=True, dir_okay=False,
                                readable=True))
def cast_metadata(paths, cast, output_file, ignore_extra,
                  error_on_missing):
    import tempfile
    from qiime2 import Metadata, metadata

    md = _merge_metadata(paths)

    cast_dict = {}
    try:
        for casting in cast:
            if ':' not in casting:
                raise click.BadParameter(
                    message=f'Missing `:` in --cast {casting}',
                    param_hint='cast')
            splitter = casting.split(':')
            if len(splitter) != 2:
                raise click.BadParameter(
                    message=f'Incorrect number of fields in --cast {casting}.'
                            f' Observed {len(splitter)}'
                            f' {tuple(splitter)}, expected 2.',
                    param_hint='cast')
            col, type_ = splitter
            if col in cast_dict:
                raise click.BadParameter(
                    message=(f'Column name "{col}" appears in cast more than'
                             ' once.'),
                    param_hint='cast')
            cast_dict[col] = type_
    except Exception as err:
        header = \
            ('Could not parse provided cast arguments into unique COLUMN:TYPE'
             ' pairs. Please make sure all cast flags are of the format --cast'
             ' COLUMN:TYPE')
        q2cli.util.exit_with_error(err, header=header)

    types = set(cast_dict.values())
    if not types.issubset(_COLUMN_TYPES):
        raise click.BadParameter(
            message=('Unknown column type provided. Please make sure all'
                     ' columns included in your cast contain a valid column'
                     ' type. Valid types: %s' %
                     (', '.join(_COLUMN_TYPES))),
            param_hint='cast')

    column_names = set(md.columns.keys())
    cast_names = set(cast_dict.keys())

    if not ignore_extra:
        if not cast_names.issubset(column_names):
            cast = cast_names.difference(column_names)
            raise click.BadParameter(
                message=('The following cast columns were not found'
                         ' within the metadata: %s' %
                         (', '.join(cast))),
                param_hint='cast')

    if error_on_missing:
        if not column_names.issubset(cast_names):
            cols = column_names.difference(cast_names)
            raise click.BadParameter(
                message='The following columns within the metadata'
                        ' were not provided in the cast: %s' %
                        (', '.join(cols)),
                param_hint='cast')

    # Remove entries from the cast dict that are not in the metadata to avoid
    # errors further down the road
    for cast in cast_names:
        if cast not in column_names:
            cast_dict.pop(cast)

    with tempfile.NamedTemporaryFile() as temp:
        md.save(temp.name)
        try:
            cast_md = Metadata.load(temp.name, cast_dict)
        except metadata.io.MetadataFileError as e:
            raise click.BadParameter(message=e, param_hint='cast') from e

    if output_file:
        cast_md.save(output_file)
    else:
        with tempfile.NamedTemporaryFile(mode='w+') as stdout_temp:
            cast_md.save(stdout_temp.name)
            stdout_str = stdout_temp.read()
            click.echo(stdout_str)


@tools.command(name='inspect-metadata',
               short_help='Inspect columns available in metadata.',
               help='Inspect metadata files or artifacts viewable as metadata.'
                    ' Providing multiple file paths to this command will merge'
                    ' the metadata.',
               cls=ToolCommand)
@click.option('--tsv/--no-tsv', default=False,
              help='Print as machine-readable TSV instead of text.')
@click.argument('paths', nargs=-1, required=True, metavar='METADATA...',
                type=click.Path(file_okay=True, dir_okay=False, readable=True))
@q2cli.util.pretty_failure(traceback=None)
def inspect_metadata(paths, tsv, failure):
    metadata = _merge_metadata(paths)

    # we aren't expecting errors below this point, so set traceback to default
    failure.traceback = 'stderr'
    failure.header = "An unexpected error has occurred:"

    COLUMN_NAME = "COLUMN NAME"
    COLUMN_TYPE = "TYPE"
    max_name_len = max([len(n) for n in metadata.columns]
                       + [len(COLUMN_NAME)])
    max_type_len = max([len(p.type) for p in metadata.columns.values()]
                       + [len(COLUMN_TYPE)])

    if tsv:
        import csv
        import io

        def formatter(*row):
            # This is gross, but less gross than robust TSV writing.
            with io.StringIO() as fh:
                writer = csv.writer(fh, dialect='excel-tab', lineterminator='')
                writer.writerow(row)
                return fh.getvalue()
    else:
        formatter = ("{0:>%d}  {1:%d}" % (max_name_len, max_type_len)).format

    click.secho(formatter(COLUMN_NAME, COLUMN_TYPE), bold=True)
    if not tsv:
        click.secho(formatter("=" * max_name_len, "=" * max_type_len),
                    bold=True)

    for name, props in metadata.columns.items():
        click.echo(formatter(name, props.type))

    if not tsv:
        click.secho(formatter("=" * max_name_len, "=" * max_type_len),
                    bold=True)
        click.secho(("{0:>%d}  " % max_name_len).format("IDS:"),
                    bold=True, nl=False)
        click.echo(metadata.id_count)
        click.secho(("{0:>%d}  " % max_name_len).format("COLUMNS:"),
                    bold=True, nl=False)
        click.echo(metadata.column_count)


def _merge_metadata(paths):
    m = [q2cli.util.load_metadata(p) for p in paths]
    metadata = m[0]
    if m[1:]:
        metadata = metadata.merge(*m[1:])

    return metadata


@tools.command(short_help='View a QIIME 2 Visualization.',
               help="Displays a QIIME 2 Visualization until the command "
                    "exits. To open a QIIME 2 Visualization so it can be "
                    "used after the command exits, use 'qiime tools extract'.",
               cls=ToolCommand)
@click.argument('visualization-path', metavar='VISUALIZATION',
                type=click.Path(file_okay=True, dir_okay=False, readable=True))
@click.option('--index-extension', required=False, default='html',
              help='The extension of the index file that should be opened. '
                   '[default: html]')
def view(visualization_path, index_extension):
    # Guard headless envs from having to import anything large
    import sys
    from qiime2 import Visualization
    from q2cli.util import _load_input
    from q2cli.core.config import CONFIG
    if not os.getenv("DISPLAY") and sys.platform != "darwin":
        raise click.UsageError(
            'Visualization viewing is currently not supported in headless '
            'environments. You can view Visualizations (and Artifacts) at '
            'https://view.qiime2.org, or move the Visualization to an '
            'environment with a display and view it with `qiime tools view`.')

    if index_extension.startswith('.'):
        index_extension = index_extension[1:]

    _, visualization = _load_input(visualization_path, view=True)[0]
    if not isinstance(visualization, Visualization):
        raise click.BadParameter(
            '%s is not a QIIME 2 Visualization. Only QIIME 2 Visualizations '
            'can be viewed.' % visualization_path)

    index_paths = visualization.get_index_paths(relative=False)

    if index_extension not in index_paths:
        raise click.BadParameter(
            'No index %s file is present in the archive. Available index '
            'extensions are: %s' % (index_extension,
                                    ', '.join(index_paths.keys())))
    else:
        index_path = index_paths[index_extension]
        launch_status = click.launch(index_path)
        if launch_status != 0:
            click.echo(CONFIG.cfg_style('error', 'Viewing visualization '
                                        'failed while attempting to open '
                                        f'{index_path}'), err=True)
        else:
            while True:
                click.echo(
                    "Press the 'q' key, Control-C, or Control-D to quit. This "
                    "view may no longer be accessible or work correctly after "
                    "quitting.", nl=False)
                # There is currently a bug in click.getchar where translation
                # of Control-C and Control-D into KeyboardInterrupt and
                # EOFError (respectively) does not work on Python 3. The code
                # here should continue to work as expected when the bug is
                # fixed in Click.
                #
                # https://github.com/pallets/click/issues/583
                try:
                    char = click.getchar()
                    click.echo()
                    if char in {'q', '\x03', '\x04'}:
                        break
                except (KeyboardInterrupt, EOFError):
                    break


@tools.command(short_help="Extract a QIIME 2 Artifact or Visualization "
                          "archive.",
               help="Extract all contents of a QIIME 2 Artifact or "
                    "Visualization's archive, including provenance, metadata, "
                    "and actual data. Use 'qiime tools export' to export only "
                    "the data stored in an Artifact or Visualization, with "
                    "the choice of exporting to different formats.",
               cls=ToolCommand)
@click.option('--input-path', required=True, metavar=_COMBO_METAVAR,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              readable=True),
              help='Path to file that should be extracted')
@click.option('--output-path', required=False,
              type=click.Path(exists=False, file_okay=False, dir_okay=True,
                              writable=True),
              help='Directory where archive should be extracted to '
                   '[default: current working directory]',
              default=os.getcwd())
def extract(input_path, output_path):
    import zipfile
    import qiime2.sdk
    from q2cli.core.config import CONFIG

    try:
        extracted_dir = qiime2.sdk.Result.extract(input_path, output_path)
    except (zipfile.BadZipFile, ValueError):
        raise click.BadParameter(
            '%s is not a valid QIIME 2 Result. Only QIIME 2 Artifacts and '
            'Visualizations can be extracted.' % input_path)
    else:
        success = 'Extracted %s to directory %s' % (input_path, extracted_dir)
        click.echo(CONFIG.cfg_style('success', success))


@tools.command(short_help='Validate data in a QIIME 2 Artifact.',
               help='Validate data in a QIIME 2 Artifact. QIIME 2 '
                    'automatically performs some basic validation when '
                    'managing your data; use this command to perform explicit '
                    'and/or more thorough validation of your data (e.g. when '
                    'debugging issues with your data or analyses).\n\nNote: '
                    'validation can take some time to complete, depending on '
                    'the size and type of your data.',
               cls=ToolCommand)
@click.argument('path', type=click.Path(exists=True, file_okay=True,
                                        dir_okay=False, readable=True),
                metavar=_COMBO_METAVAR)
@click.option('--level', required=False, type=click.Choice(['min', 'max']),
              help='Desired level of validation. "min" will perform minimal '
                   'validation, and "max" will perform maximal validation (at '
                   'the potential cost of runtime).',
              default='max', show_default=True)
def validate(path, level):
    import qiime2.sdk
    from q2cli.core.config import CONFIG

    try:
        result = qiime2.sdk.Result.load(path)
    except Exception as e:
        header = 'There was a problem loading %s as a QIIME 2 Result:' % path
        q2cli.util.exit_with_error(e, header=header)

    try:
        result.validate(level)
    except qiime2.plugin.ValidationError as e:
        header = 'Result %s does not appear to be valid at level=%s:' % (
                path, level)
        q2cli.util.exit_with_error(e, header=header, traceback=None)
    except Exception as e:
        header = ('An unexpected error has occurred while attempting to '
                  'validate result %s:' % path)
        q2cli.util.exit_with_error(e, header=header)
    else:
        click.echo(CONFIG.cfg_style('success', f'Result {path} appears to be '
                                    f'valid at level={level}.'))


@tools.command(short_help='Print citations for a QIIME 2 result.',
               help='Print citations as a BibTex file (.bib) for a QIIME 2'
                    ' result.',
               cls=ToolCommand)
@click.argument('path', type=click.Path(exists=True, file_okay=True,
                                        dir_okay=False, readable=True),
                metavar=_COMBO_METAVAR)
def citations(path):
    import qiime2.sdk
    import io
    from q2cli.core.config import CONFIG
    ctx = click.get_current_context()

    try:
        result = qiime2.sdk.Result.load(path)
    except Exception as e:
        header = 'There was a problem loading %s as a QIIME 2 result:' % path
        q2cli.util.exit_with_error(e, header=header)

    if result.citations:
        with io.StringIO() as fh:
            result.citations.save(fh)
            click.echo(fh.getvalue(), nl=False)
        ctx.exit(0)
    else:
        click.echo(CONFIG.cfg_style('problem', 'No citations found.'),
                   err=True)
        ctx.exit(1)


@tools.command(name='cache-create',
               short_help='Create an empty cache at the given location.',
               help='Create an empty cache at the given location.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=False, readable=True),
              help='Path to a nonexistent directory to be created as a cache.')
def cache_create(cache):
    from qiime2.core.cache import Cache
    from q2cli.core.config import CONFIG

    try:
        Cache(cache)
    except Exception as e:
        header = "There was a problem creating a cache at '%s':" % cache
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    success = "Created cache at '%s'" % cache
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(name='cache-remove',
               short_help='Removes a given key from a cache.',
               help='Removes a given key from a cache then runs garbage '
                    'collection on the cache.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to remove the key from.')
@click.option('--key', required=True,
              help='The key to remove from the cache.')
def cache_remove(cache, key):
    from qiime2.core.cache import Cache
    from q2cli.core.config import CONFIG

    try:
        _cache = Cache(cache)
        _cache.remove(key)
    except Exception as e:
        header = "There was a problem removing the key '%s' from the " \
                 "cache '%s':" % (key, cache)
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    success = "Removed key '%s' from cache '%s'" % (key, cache)
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(name='cache-garbage-collection',
               short_help='Runs garbage collection on the cache at the '
                          'specified location.',
               help='Runs garbage collection on the cache at the specified '
                    'location if the specified location is a cache.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to run garbage collection on.')
def cache_garbage_collection(cache):
    from qiime2.core.cache import Cache
    from q2cli.core.config import CONFIG

    try:
        _cache = Cache(cache)
        _cache.garbage_collection()
    except Exception as e:
        header = "There was a problem running garbage collection on the " \
            "cache at '%s':" % cache
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    success = "Ran garbage collection on cache at '%s'" % cache
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(name='cache-store',
               short_help='Stores a .qza in the cache under a key.',
               help='Stores a .qza in the cache under a key.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to save into.')
@click.option('--artifact-path', required=True,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              readable=True),
              help='Path to a .qza to save into the cache.')
@click.option('--key', required=True,
              help='The key to save the artifact under (must be a valid '
                   'Python identifier).')
def cache_store(cache, artifact_path, key):
    from qiime2.sdk.result import Result
    from qiime2.core.cache import Cache
    from q2cli.core.config import CONFIG

    try:
        artifact = Result.load(artifact_path)
        _cache = Cache(cache)
        _cache.save(artifact, key)
    except Exception as e:
        header = "There was a problem saving the artifact '%s' to the cache " \
                 "'%s' under the key '%s':" % (artifact_path, cache, key)
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    success = "Saved the artifact '%s' to the cache '%s' under the key " \
        "'%s'" % (artifact_path, cache, key)
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(name='cache-import',
               short_help='Imports data into an Artifact in the cache under a '
                          'key.',
               help='Imports data into an Artifact in the cache under a key.',
               cls=ToolCommand)
@click.option('--type', required=True,
              help='The semantic type of the artifact that will be created '
                   'upon importing. Use `qiime tools list-types` to see what '
                   'importable semantic types are available in the current '
                   'deployment.')
@click.option('--input-path', required=True,
              type=click.Path(exists=True, file_okay=True, dir_okay=True,
                              readable=True),
              help='Path to file or directory that should be imported.')
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to save into.')
@click.option('--key', required=True,
              help='The key to save the artifact under (must be a valid '
                   'Python identifier).')
@click.option('--input-format', required=False,
              help='The format of the data to be imported. If not provided, '
                   'data must be in the format expected by the semantic type '
                   'provided via --type. Use `qiime tools list-formats '
                   '--importable` to see which formats of input data are '
                   'importable.')
@click.option('--validate-level', required=False, default='max',
              type=click.Choice(['min', 'max']),
              help='How much to validate the imported data before creating the'
                   ' artifact. A value of "max" will generally read the entire'
                   ' file or directory, whereas "min" will not usually do so.'
                   ' [default: "max"]')
def cache_import(type, input_path, cache, key, input_format, validate_level):
    from qiime2 import Cache
    from q2cli.core.config import CONFIG

    artifact = _import(type, input_path, input_format, validate_level)
    _cache = Cache(cache)
    _cache.save(artifact, key)

    if input_format is None:
        input_format = artifact.format.__name__

    success = 'Imported %s as %s to %s:%s' % (input_path,
                                              input_format,
                                              cache,
                                              key)
    click.echo(CONFIG.cfg_style('success', success))


def _import(type, input_path, input_format, validate_level):
    import qiime2.sdk
    import qiime2.plugin

    try:
        artifact = qiime2.sdk.Artifact.import_data(
            type, input_path, view_type=input_format,
            validate_level=validate_level)
    except qiime2.plugin.ValidationError as e:
        header = 'There was a problem importing %s:' % input_path
        q2cli.util.exit_with_error(e, header=header, traceback=None)
    except Exception as e:
        header = 'An unexpected error has occurred:'
        q2cli.util.exit_with_error(e, header=header)

    return artifact


@tools.command(name='cache-fetch',
               short_help='Fetches an artifact out of a cache into a .qza.',
               help='Fetches the artifact saved to the specified cache under '
                    'the specified key into a .qza at the specified location.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to load from.')
@click.option('--key', required=True,
              help='The key to the artifact being loaded.')
@click.option('--output-path', required=True,
              type=click.Path(exists=False, readable=True),
              help='Path to put the .qza we are loading the artifact into.')
def cache_fetch(cache, key, output_path):
    from qiime2.core.cache import Cache
    from q2cli.core.config import CONFIG

    try:
        _cache = Cache(cache)
        artifact = _cache.load(key)
        artifact.save(output_path)
    except Exception as e:
        header = "There was a problem loading the artifact with the key " \
                 "'%s' from the cache '%s' and saving it to the file '%s':" % \
                 key, cache, output_path
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    success = "Loaded artifact with the key '%s' from the cache '%s' and " \
        "saved it to the file '%s'" % (key, cache, output_path)
    click.echo(CONFIG.cfg_style('success', success))


@tools.command(name='cache-status',
               short_help='Checks the status of the cache.',
               help='Lists all keys in the given cache. Peeks artifacts '
                    'pointed to by keys to data and lists the number of '
                    'artifacts in the pool for keys to pools.',
               cls=ToolCommand)
@click.option('--cache', required=True,
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              readable=True),
              help='Path to an existing cache to check the status of.')
def cache_status(cache):
    from qiime2.core.cache import Cache
    from qiime2.sdk.result import Result

    from q2cli.core.config import CONFIG

    data_output = []
    pool_output = []
    try:
        _cache = Cache(cache)
        with _cache.lock:
            for key in _cache.get_keys():
                key_values = _cache.read_key(key)

                if 'data' in key_values:
                    data = key_values['data']
                    data_output.append(
                        'data: %s -> %s' %
                        (key, str(Result.peek(_cache.data / data))))
                elif 'pool' in key_values:
                    pool = key_values['pool']
                    pool_output.append(
                        'pool: %s -> size = %s' %
                        (key, str(len(os.listdir(_cache.pools / pool)))))
    except Exception as e:
        header = "There was a problem getting the status of the cache at " \
                 "path '%s':" % cache
        q2cli.util.exit_with_error(e, header=header, traceback=None)

    if not data_output:
        data_output = 'No data keys in cache'
    else:
        data_output = '\n'.join(data_output)
        data_output = 'Data keys in cache:\n' + data_output

    if not pool_output:
        pool_output = 'No pool keys in cache'
    else:
        pool_output = '\n'.join(pool_output)
        pool_output = 'Pool keys in cache:\n' + pool_output

    output = data_output + '\n\n' + pool_output
    success = "Status of the cache at the path '%s':\n\n%s" % \
        (cache, output)
    click.echo(CONFIG.cfg_style('success', success))


replay_in_fp_help = (
    'filepath to a QIIME 2 Archive (.qza or .qzv) or directory of Archives'
)
replay_recurse_help = (
    'if in-fp is a directory, will also search sub-directories when finding '
    'Archives to parse'
)
replay_validate_checksums_help = (
    'check that replayed archives are intact and uncorrupted'
)
replay_parse_metadata_help = (
    'parse the original metadata captured in provenance for review or replay'
)
replay_use_recorded_metadata_help = (
    're-use the original metadata captured in provenance'
)
replay_suppress_header_help = (
   'do not write header/footer blocks in the output script'
)
replay_verbose_help = (
    'print status messages to stdout while processing'
)
replay_dump_recorded_metadata_help = (
    'write the original metadata captured in provenance to disk in the '
    '--metadata-out-dir directory'
)


@tools.command(name='replay-provenance', cls=ToolCommand)
@click.option('--in-fp', required=True, help=replay_in_fp_help)
@click.option('--recurse/--no-recurse',
              default=False,
              show_default=True,
              help=replay_recurse_help)
@click.option('--usage-driver',
              default='cli',
              show_default=True,
              help='the target interface for your replay script',
              type=click.Choice(['python3', 'cli'], case_sensitive=False))
@click.option('--validate-checksums/--no-validate-checksums',
              default=True,
              show_default=True,
              help=replay_validate_checksums_help)
@click.option('--parse-metadata/--no-parse-metadata',
              default=True,
              show_default=True,
              help=replay_parse_metadata_help)
@click.option('--use-recorded-metadata/--no-use-recorded-metadata',
              default=False,
              show_default=True,
              help=replay_use_recorded_metadata_help)
@click.option('--suppress-header/--no-suppress-header',
              default=False,
              show_default=True,
              help=replay_suppress_header_help)
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help=replay_verbose_help)
@click.option('--dump-recorded-metadata/--no-dump-recorded-metadata',
              default=True,
              show_default=True,
              help=replay_dump_recorded_metadata_help)
@click.option('--metadata-out-dir',
              default='',
              show_default=True,
              help=('the directory where captured study metadata '
                    'should be written if --dump-recorded-metadata. This '
                    'often produces many outputs, so a dedicated directory '
                    'should generally be used. Creates the directory if it '
                    'does not already exist. By default, metadata is written '
                    'to `${PWD}/recorded_metadata/`'))
@click.option('--out-fp',
              required=True,
              type=click.Path(exists=False, writable=True),
              help='the filepath where your replay script should be written')
def provenance_replay(
    in_fp: str,
    out_fp: str,
    usage_driver: Literal['python3', 'cli'],
    recurse: bool = False,
    validate_checksums: bool = True,
    parse_metadata: bool = True,
    use_recorded_metadata: bool = False,
    suppress_header: bool = False,
    verbose: bool = True,
    dump_recorded_metadata: bool = True,
    metadata_out_dir: str = ''
):
    """
    Replay provenance from a QIIME 2 Artifact filepath to a written executable
    """
    from qiime2.core.archive.provenance_lib.replay import replay_provenance
    from qiime2.sdk.util import get_available_usage_drivers

    usage_drivers = get_available_usage_drivers()
    try:
        usage_driver_type = usage_drivers[usage_driver]
    except KeyError:
        msg = (
            f'The {usage_driver} usage driver is not available in the '
            'current environment.'
        )
        raise ValueError(msg)

    replay_provenance(
        usage_driver=usage_driver_type,
        payload=in_fp,
        out_fp=out_fp,
        validate_checksums=validate_checksums,
        parse_metadata=parse_metadata,
        recurse=recurse,
        use_recorded_metadata=use_recorded_metadata,
        suppress_header=suppress_header,
        verbose=verbose,
        dump_recorded_metadata=dump_recorded_metadata,
        md_out_dir=metadata_out_dir
    )
    filename = os.path.realpath(out_fp)
    click.echo(f'{usage_driver} replay script written to {filename}')


@tools.command(name='replay-citations', cls=ToolCommand)
@click.option('--in-fp', required=True, help=replay_in_fp_help)
@click.option('--recurse/--no-recurse',
              default=False,
              show_default=True,
              help=replay_recurse_help)
@click.option('--deduplicate/--no-deduplicate',
              default=True,
              show_default=True,
              help=('If deduplicate, duplicate citations will be removed '
                    'heuristically, e.g. by comparing DOI fields. '
                    'This greatly reduces manual curation of reference lists, '
                    'but introduces a small risk of reference loss.'))
@click.option('--suppress-header/--no-suppress-header',
              default=False,
              show_default=True,
              help=replay_suppress_header_help)
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help=replay_verbose_help)
@click.option('--out-fp',
              required=True,
              type=click.Path(exists=False, writable=True),
              help='the filepath where your bibtex file should be written')
def citations_replay(
    in_fp: str,
    out_fp: str,
    recurse: bool = False,
    deduplicate: bool = True,
    suppress_header: bool = False,
    verbose: bool = True
):
    """
    Reports all citations from a QIIME 2 Artifact or directory of Artifacts,
    with the goal of improving and simplifying attribution of/in published
    work.

    Not for use in reporting e.g. software versions used in an analysis, as
    deduplication removes duplicate references with different plugin versions.
    """
    from qiime2.core.archive.provenance_lib.parse import ProvDAG
    from qiime2.core.archive.provenance_lib.replay import replay_citations

    dag = ProvDAG(in_fp, verbose=verbose, recurse=recurse)
    replay_citations(
        dag,
        out_fp=out_fp,
        deduplicate=deduplicate,
        suppress_header=suppress_header
    )
    filename = os.path.realpath(out_fp)
    click.echo(f'citations bibtex file written to {filename}')


@tools.command(name='replay-supplement', cls=ToolCommand)
@click.option('--in-fp', required=True,
              help='filepath to a QIIME 2 Archive or directory of Archives')
@click.option('--recurse/--no-recurse',
              default=False,
              show_default=True,
              help=('if in-fp is a directory, will also search sub-directories'
                    ' when finding .qza/.qzv files to parse'))
@click.option('--deduplicate/--no-deduplicate',
              default=True,
              show_default=True,
              help=('If deduplicate, duplicate citations will be removed '
                    'heuristically, e.g. by comparing DOI fields. '
                    'This greatly reduces manual curation of reference lists, '
                    'but introduces a small risk of reference loss.'))
@click.option('--validate-checksums/--no-validate-checksums',
              default=True,
              show_default=True,
              help=replay_validate_checksums_help)
@click.option('--parse-metadata/--no-parse-metadata',
              default=True,
              show_default=True,
              help=replay_parse_metadata_help)
@click.option('--use-recorded-metadata/--no-use-recorded-metadata',
              default=False,
              show_default=True,
              help=replay_use_recorded_metadata_help)
@click.option('--suppress-header/--no-suppress-header',
              default=False,
              show_default=True,
              help=replay_suppress_header_help)
@click.option('--verbose/--no-verbose',
              default=True,
              show_default=True,
              help=replay_verbose_help)
@click.option('--dump-recorded-metadata/--no-dump-recorded-metadata',
              default=True,
              show_default=True,
              help='write the original metadata captured in provenance to '
                   'recorded_metadata/ inside the archive')
@click.option('--out-fp',
              required=True,
              type=click.Path(exists=False, writable=True),
              help='the filepath where your reproduciblity supplement zipfile '
                   'should be written')
def supplement_replay(
    in_fp: str,
    out_fp: str,
    validate_checksums: bool = True,
    parse_metadata: bool = True,
    use_recorded_metadata: bool = False,
    recurse: bool = False,
    deduplicate: bool = True,
    suppress_header: bool = False,
    verbose: bool = True,
    dump_recorded_metadata: bool = True
):
    """
    Produces a zipfile package of useful documentation supporting in silico
    reproducibility of some QIIME 2 Result(s) from a QIIME 2 Artifact or
    directory of Artifacts.

    Package includes:
    - replay scripts for all supported interfaces
    - a bibtex-formatted collection of all citations
    """
    from qiime2.core.archive.provenance_lib.replay import replay_supplement
    from qiime2.sdk.util import get_available_usage_drivers

    usage_drivers = get_available_usage_drivers()
    usage_driver_types = list(usage_drivers.values())
    if not usage_driver_types:
        msg = (
            'There are no available usage drivers registered in the current '
            'environment.'
        )
        raise ValueError(msg)

    replay_supplement(
        usage_drivers=usage_driver_types,
        payload=in_fp,
        out_fp=out_fp,
        validate_checksums=validate_checksums,
        parse_metadata=parse_metadata,
        use_recorded_metadata=use_recorded_metadata,
        recurse=recurse,
        deduplicate=deduplicate,
        suppress_header=suppress_header,
        verbose=verbose,
        dump_recorded_metadata=dump_recorded_metadata
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from q2cli.click.command import ToolCommand


def _echo_version():
    import sys
    import qiime2
    import q2cli

    pyver = sys.version_info
    click.echo('Python version: %d.%d.%d' %
               (pyver.major, pyver.minor, pyver.micro))
    click.echo('QIIME 2 release: %s' % qiime2.__release__)
    click.echo('QIIME 2 version: %s' % qiime2.__version__)
    click.echo('q2cli version: %s' % q2cli.__version__)


def _echo_plugins():
    import q2cli.core.cache

    plugins = q2cli.core.cache.CACHE.plugins
    if plugins:
        for name, plugin in sorted(plugins.items()):
            click.echo('%s: %s' % (name, plugin['version']))
    else:
        click.secho('No plugins are currently installed.\nYou can browse '
                    'the official QIIME 2 plugins at https://qiime2.org')


@click.command(help='Display information about current deployment.',
               cls=ToolCommand)
def info():
    import q2cli.util
    # This import improves performance for repeated _echo_plugins
    import q2cli.core.cache

    click.secho('System versions', fg='green')
    _echo_version()
    click.secho('\nInstalled plugins', fg='green')
    _echo_plugins()

    click.secho('\nApplication config directory', fg='green')
    click.secho(q2cli.util.get_app_dir())

    click.secho('\nGetting help', fg='green')
    click.secho('To get help with QIIME 2, visit https://qiime2.org')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import ast
import shutil
import os.path
import unittest
import unittest.mock
import tempfile
import pkg_resources

from click.testing import CliRunner
from qiime2.core.testing.type import (IntSequence1, IntSequence2, Mapping,
                                      SingleInt)
from qiime2.core.testing.util import get_dummy_plugin
from qiime2.core.util import load_action_yaml
from qiime2.core.cache import Cache

from q2cli.commands import RootCommand
from q2cli.builtin.tools import tools
from q2cli.util import get_default_recycle_pool
from qiime2.sdk import Artifact, Visualization, ResultCollection
from qiime2.sdk.parallel_config import PARALLEL_CONFIG


# What to split the errors raised by intentionally failed pipeline on to get
# at the uuids needed for testing
FIRST_SPLIT = 'Plugin error from dummy-plugin:\n\n  '
SECOND_SPLIT = '\n\nSee above for debug info.'


def get_data_path(filename):
    return pkg_resources.resource_filename('q2cli.tests', 'data/%s' % filename)


class TestCacheCli(unittest.TestCase):
    def setUp(self):
        get_dummy_plugin()

        self.runner = CliRunner()
        self.plugin_command = RootCommand().get_command(
            ctx=None, name='dummy-plugin')
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')
        self.cache = Cache(os.path.join(self.tempdir, 'new_cache'))

        self.art1 = Artifact.import_data(IntSequence1, [0, 1, 2])
        self.art2 = Artifact.import_data(IntSequence1, [3, 4, 5])
        self.art3 = Artifact.import_data(IntSequence2, [6, 7, 8])
        self.art4 = Artifact.import_data(SingleInt, 0)
        self.art5 = Artifact.import_data(SingleInt, 1)
        self.ints1 = {'1': self.art4, '2': self.art5}
        self.ints2 = {'1': self.art1, '2': self.art2}
        self.mapping = Artifact.import_data(Mapping, {'a': '1', 'b': '2'})

        self.metadata = os.path.join(self.tempdir, 'metadata.tsv')
        with open(self.metadata, 'w') as fh:
            fh.write('#SampleID\tcol1\n0\tfoo\nid1\tbar\n')

        self.non_cache_output = os.path.join(self.tempdir, 'output.qza')
        self.art3_non_cache = os.path.join(self.tempdir, 'art3.qza')

        # Ensure default state prior to test
        PARALLEL_CONFIG.parallel_config = None
        PARALLEL_CONFIG.action_executor_mapping = {}

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _run_command(self, *args):
        return self.runner.invoke(self.plugin_command, args)

    def test_inputs_from_cache(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', self.non_cache_output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(Artifact.load(self.non_cache_output).view(list),
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_inputs_split(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.art3.save(self.art3_non_cache)

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', self.art3_non_cache, '--p-int1', '9', '--p-int2',
            '10', '--o-concatenated-ints', self.non_cache_output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(Artifact.load(self.non_cache_output).view(list),
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_colon_in_input_path_not_cache(self):
        art_path = os.path.join(self.tempdir, 'art:1.qza')
        self.art1.save(art_path)

        left_path = os.path.join(self.tempdir, 'left.qza')
        right_path = os.path.join(self.tempdir, 'right.qza')

        result = self._run_command(
            'split-ints', '--i-ints', art_path, '--o-left', left_path,
            '--o-right', right_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(Artifact.load(left_path).view(list), [0])
        self.assertEqual(Artifact.load(right_path).view(list), [1, 2])

    def test_colon_in_cache_path(self):
        cache = Cache(os.path.join(self.tempdir, 'new:cache'))
        cache.save(self.art1, 'art')

        art_path = str(cache.path) + ':art'

        left_path = os.path.join(self.tempdir, 'left.qza')
        right_path = os.path.join(self.tempdir, 'right.qza')

        result = self._run_command(
            'split-ints', '--i-ints', art_path, '--o-left', left_path,
            '--o-right', right_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(Artifact.load(left_path).view(list), [0])
        self.assertEqual(Artifact.load(right_path).view(list), [1, 2])

    def test_output_to_cache(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        out_path = str(self.cache.path) + ':out'

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', out_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.cache.load('out').view(list),
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_outputs_to_cache(self):
        self.cache.save(self.art1, 'art1')
        art1_path = str(self.cache.path) + ':art1'

        left_path = str(self.cache.path) + ':left'
        right_path = str(self.cache.path) + ':right'

        result = self._run_command(
            'split-ints', '--i-ints', art1_path, '--o-left', left_path,
            '--o-right', right_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.cache.load('left').view(list), [0])
        self.assertEqual(self.cache.load('right').view(list), [1, 2])

    def test_outputs_split(self):
        self.cache.save(self.art1, 'art1')
        art1_path = str(self.cache.path) + ':art1'

        left_path = str(self.cache.path) + ':left'

        result = self._run_command(
            'split-ints', '--i-ints', art1_path, '--o-left', left_path,
            '--o-right', self.non_cache_output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.cache.load('left').view(list), [0])
        self.assertEqual(Artifact.load(self.non_cache_output).view(list),
                         [1, 2])

    def test_invalid_cache_path_input(self):
        art1_path = 'not_a_cache:art1'

        left_path = str(self.cache.path) + ':left'
        right_path = str(self.cache.path) + ':right'

        result = self._run_command(
            'split-ints', '--i-ints', art1_path, '--o-left', left_path,
            '--o-right', right_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.output, r"cache")

    def test_invalid_cache_path_output(self):
        self.cache.save(self.art1, 'art1')
        art1_path = str(self.cache.path) + ':art1'

        left_path = '/this/is/not_a_cache:left'
        right_path = str(self.cache.path) + ':right'

        result = self._run_command(
            'split-ints', '--i-ints', art1_path, '--o-left', left_path,
            '--o-right', right_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('does not exist', result.output)

    def test_colon_in_out_path_not_cache(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        out_path = os.path.join(self.tempdir, 'out:put.qza')

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', out_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(Artifact.load(out_path).view(list),
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_collection_roundtrip_list(self):
        key1 = 'out1'
        key2 = 'out2'

        collection_out1 = str(self.cache.path) + ':' + key1
        collection_out2 = str(self.cache.path) + ':' + key2

        result = self._run_command(
            'list-params', '--p-ints', '0', '--p-ints', '1', '--o-output',
            collection_out1, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key1)

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

        result = self._run_command(
            'list-of-ints', '--i-ints', collection_out1, '--o-output',
            collection_out2, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key2)

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

    def test_collection_roundtrip_dict_keyed(self):
        key1 = 'out1'
        key2 = 'out2'

        collection_out1 = str(self.cache.path) + ':' + key1
        collection_out2 = str(self.cache.path) + ':' + key2

        result = self._run_command(
            'dict-params', '--p-ints', 'foo:0', '--p-ints', 'bar:1',
            '--o-output', collection_out1, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key1)

        self.assertEqual(collection['foo'].view(int), 0)
        self.assertEqual(collection['bar'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['foo', 'bar'])

        result = self._run_command(
            'dict-of-ints', '--i-ints', collection_out1, '--o-output',
            collection_out2, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key2)

        self.assertEqual(collection['foo'].view(int), 0)
        self.assertEqual(collection['bar'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['foo', 'bar'])

    def test_collection_roundtrip_dict_unkeyed(self):
        key1 = 'out1'
        key2 = 'out2'

        collection_out1 = str(self.cache.path) + ':' + key1
        collection_out2 = str(self.cache.path) + ':' + key2

        result = self._run_command(
            'dict-params', '--p-ints', '0', '--p-ints', '1',
            '--o-output', collection_out1, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key1)

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

        result = self._run_command(
            'dict-of-ints', '--i-ints', collection_out1, '--o-output',
            collection_out2, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection(key2)

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

    def test_de_facto_list(self):
        self.cache.save(self.art4, 'art4')
        self.cache.save(self.art5, 'art5')

        art4_path = str(self.cache.path) + ':art4'
        art5_path = str(self.cache.path) + ':art5'
        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'list-of-ints', '--i-ints', art4_path, '--i-ints', art5_path,
            '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection('output')

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

    def test_de_facto_dict_keyed(self):
        self.cache.save(self.art4, 'art4')
        self.cache.save(self.art5, 'art5')

        art4_path = str(self.cache.path) + ':art4'
        art5_path = str(self.cache.path) + ':art5'
        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'dict-of-ints', '--i-ints', f'foo:{art4_path}', '--i-ints',
            f'bar:{art5_path}', '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection('output')

        self.assertEqual(collection['foo'].view(int), 0)
        self.assertEqual(collection['bar'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['foo', 'bar'])

    def test_de_facto_dict_unkeyed(self):
        self.cache.save(self.art4, 'art4')
        self.cache.save(self.art5, 'art5')

        art4_path = str(self.cache.path) + ':art4'
        art5_path = str(self.cache.path) + ':art5'
        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'dict-of-ints', '--i-ints', art4_path, '--i-ints', art5_path,
            '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection('output')

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

    def test_mixed_cached_uncached_inputs(self):
        art4_path = os.path.join(self.tempdir, 'art4.qza')
        self.art4.save(art4_path)

        self.cache.save(self.art5, 'art5')
        art5_path = str(self.cache.path) + ':art5'

        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'dict-of-ints', '--i-ints', art4_path, '--i-ints',
            art5_path, '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection('output')

        self.assertEqual(collection['0'].view(int), 0)
        self.assertEqual(collection['1'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['0', '1'])

        self.cache.remove('output')

        result = self._run_command(
            'dict-of-ints', '--i-ints', f'foo:{art4_path}', '--i-ints',
            f'bar:{art5_path}', '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 0)
        collection = self.cache.load_collection('output')

        self.assertEqual(collection['foo'].view(int), 0)
        self.assertEqual(collection['bar'].view(int), 1)
        self.assertEqual(list(collection.keys()), ['foo', 'bar'])

    def test_pipeline_resumption_default(self):
        plugin_action = 'dummy_plugin_resumable_varied_pipeline'
        default_pool = get_default_recycle_pool(plugin_action)
        default_pool_fp = os.path.join(self.cache.pools, default_pool)
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        self.cache.save_collection(self.ints2, 'ints2')
        self.cache.save(self.art4, 'int1')

        ints1_path = str(self.cache.path) + ':ints1'
        ints2_path = str(self.cache.path) + ':ints2'
        int1_path = str(self.cache.path) + ':int1'

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--p-fail', 'True',
            '--output-dir', output, '--use-cache', str(self.cache.path),
            '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        # Assert that the pool exists
        self.assertTrue(os.path.exists(default_pool_fp))

        exception = result.output.split(FIRST_SPLIT)[-1]
        exception = exception.split(SECOND_SPLIT)[0]

        ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
            identity_uuid, viz_uuid = ast.literal_eval(exception)

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--output-dir', output,
            '--use-cache', str(self.cache.path), '--verbose'
        )

        self.assertEqual(result.exit_code, 0)

        ints1_ret_fp = os.path.join(output, 'ints1_ret')
        ints2_ret_fp = os.path.join(output, 'ints2_ret')
        int1_ret_fp = os.path.join(output, 'int1_ret.qza')
        list_ret_fp = os.path.join(output, 'list_ret')
        dict_ret_fp = os.path.join(output, 'dict_ret')
        identity_ret_fp = os.path.join(output, 'identity_ret.qza')
        viz_ret_fp = os.path.join(output, 'viz.qzv')

        ints1_ret = ResultCollection.load(ints1_ret_fp)
        ints2_ret = ResultCollection.load(ints2_ret_fp)
        int1_ret = Artifact.load(int1_ret_fp)
        list_ret = ResultCollection.load(list_ret_fp)
        dict_ret = ResultCollection.load(dict_ret_fp)
        identity_ret = Artifact.load(identity_ret_fp)
        viz_ret = Visualization.load(viz_ret_fp)

        complete_ints1_uuids = self._load_alias_uuids(ints1_ret)
        complete_ints2_uuids = self._load_alias_uuids(ints2_ret)
        complete_int1_uuid = self._load_alias_uuid(int1_ret)
        complete_list_uuids = self._load_alias_uuids(list_ret)
        complete_dict_uuids = self._load_alias_uuids(dict_ret)
        complete_identity_uuid = self._load_alias_uuid(identity_ret)
        complete_viz_uuid = self._load_alias_uuid(viz_ret)

        # Assert that the artifacts returned by the completed run of the
        # pipeline are aliases of the artifacts created by the first failed run
        self.assertEqual(ints1_uuids, complete_ints1_uuids)
        self.assertEqual(ints2_uuids, complete_ints2_uuids)
        self.assertEqual(int1_uuid, complete_int1_uuid)
        self.assertEqual(list_uuids, complete_list_uuids)
        self.assertEqual(dict_uuids, complete_dict_uuids)
        self.assertEqual(identity_uuid, complete_identity_uuid)
        self.assertEqual(viz_uuid, complete_viz_uuid)

        # Assert that the pool was removed
        self.assertFalse(os.path.exists(default_pool_fp))

    def test_pipeline_resumption_different_pool(self):
        pool = 'pool'
        pool_fp = os.path.join(self.cache.pools, pool)
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        self.cache.save_collection(self.ints2, 'ints2')
        self.cache.save(self.art4, 'int1')

        ints1_path = str(self.cache.path) + ':ints1'
        ints2_path = str(self.cache.path) + ':ints2'
        int1_path = str(self.cache.path) + ':int1'

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--p-fail', 'True',
            '--output-dir', output, '--recycle-pool', pool, '--use-cache',
            str(self.cache.path), '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        # Assert that the pool exists
        self.assertTrue(os.path.exists(pool_fp))

        exception = result.output.split(FIRST_SPLIT)[-1]
        exception = exception.split(SECOND_SPLIT)[0]

        ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
            identity_uuid, viz_uuid = ast.literal_eval(exception)

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--output-dir', output,
            '--recycle-pool', pool, '--use-cache', str(self.cache.path),
            '--verbose'
        )

        self.assertEqual(result.exit_code, 0)

        ints1_ret_fp = os.path.join(output, 'ints1_ret')
        ints2_ret_fp = os.path.join(output, 'ints2_ret')
        int1_ret_fp = os.path.join(output, 'int1_ret.qza')
        list_ret_fp = os.path.join(output, 'list_ret')
        dict_ret_fp = os.path.join(output, 'dict_ret')
        identity_ret_fp = os.path.join(output, 'identity_ret.qza')
        viz_ret_fp = os.path.join(output, 'viz.qzv')

        ints1_ret = ResultCollection.load(ints1_ret_fp)
        ints2_ret = ResultCollection.load(ints2_ret_fp)
        int1_ret = Artifact.load(int1_ret_fp)
        list_ret = ResultCollection.load(list_ret_fp)
        dict_ret = ResultCollection.load(dict_ret_fp)
        identity_ret = Artifact.load(identity_ret_fp)
        viz_ret = Visualization.load(viz_ret_fp)

        complete_ints1_uuids = self._load_alias_uuids(ints1_ret)
        complete_ints2_uuids = self._load_alias_uuids(ints2_ret)
        complete_int1_uuid = self._load_alias_uuid(int1_ret)
        complete_list_uuids = self._load_alias_uuids(list_ret)
        complete_dict_uuids = self._load_alias_uuids(dict_ret)
        complete_identity_uuid = self._load_alias_uuid(identity_ret)
        complete_viz_uuid = self._load_alias_uuid(viz_ret)

        # Assert that the artifacts returned by the completed run of the
        # pipeline are aliases of the artifacts created by the first failed run
        self.assertEqual(ints1_uuids, complete_ints1_uuids)
        self.assertEqual(ints2_uuids, complete_ints2_uuids)
        self.assertEqual(int1_uuid, complete_int1_uuid)
        self.assertEqual(list_uuids, complete_list_uuids)
        self.assertEqual(dict_uuids, complete_dict_uuids)
        self.assertEqual(identity_uuid, complete_identity_uuid)
        self.assertEqual(viz_uuid, complete_viz_uuid)

        # Assert that the pool is still there
        self.assertTrue(os.path.exists(pool_fp))

    def test_pipeline_resumption_no_recycle(self):
        plugin_action = 'dummy_plugin_resumable_varied_pipeline'
        default_pool = get_default_recycle_pool(plugin_action)
        default_pool_fp = os.path.join(self.cache.pools, default_pool)
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        self.cache.save_collection(self.ints2, 'ints2')
        self.cache.save(self.art4, 'int1')

        ints1_path = str(self.cache.path) + ':ints1'
        ints2_path = str(self.cache.path) + ':ints2'
        int1_path = str(self.cache.path) + ':int1'

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--p-fail', 'True',
            '--output-dir', output, '--use-cache', str(self.cache.path),
            '--no-recycle', '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        # Assert that the pool was not created
        self.assertFalse(os.path.exists(default_pool_fp))

        exception = result.output.split(FIRST_SPLIT)[-1]
        exception = exception.split(SECOND_SPLIT)[0]

        ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
            identity_uuid, viz_uuid = ast.literal_eval(exception)

        result = self._run_command(
            'resumable-varied-pipeline', '--i-ints1', ints1_path, '--i-ints2',
            ints2_path, '--i-int1', int1_path, '--p-string', 'Hi',
            '--m-metadata-file', self.metadata, '--output-dir', output,
            '--use-cache', str(self.cache.path), '--verbose'
        )

        self.assertEqual(result.exit_code, 0)

        ints1_ret_fp = os.path.join(output, 'ints1_ret')
        ints2_ret_fp = os.path.join(output, 'ints2_ret')
        int1_ret_fp = os.path.join(output, 'int1_ret.qza')
        list_ret_fp = os.path.join(output, 'list_ret')
        dict_ret_fp = os.path.join(output, 'dict_ret')
        identity_ret_fp = os.path.join(output, 'identity_ret.qza')
        viz_ret_fp = os.path.join(output, 'viz.qzv')

        ints1_ret = ResultCollection.load(ints1_ret_fp)
        ints2_ret = ResultCollection.load(ints2_ret_fp)
        int1_ret = Artifact.load(int1_ret_fp)
        list_ret = ResultCollection.load(list_ret_fp)
        dict_ret = ResultCollection.load(dict_ret_fp)
        identity_ret = Artifact.load(identity_ret_fp)
        viz_ret = Visualization.load(viz_ret_fp)

        complete_ints1_uuids = self._load_alias_uuids(ints1_ret)
        complete_ints2_uuids = self._load_alias_uuids(ints2_ret)
        complete_int1_uuid = self._load_alias_uuid(int1_ret)
        complete_list_uuids = self._load_alias_uuids(list_ret)
        complete_dict_uuids = self._load_alias_uuids(dict_ret)
        complete_identity_uuid = self._load_alias_uuid(identity_ret)
        complete_viz_uuid = self._load_alias_uuid(viz_ret)

        # Assert that the artifacts returned by the completed run of the
        # pipeline are aliases of the artifacts created by the first failed run
        self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
        self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
        self.assertNotEqual(int1_uuid, complete_int1_uuid)
        self.assertNotEqual(list_uuids, complete_list_uuids)
        self.assertNotEqual(dict_uuids, complete_dict_uuids)
        self.assertNotEqual(identity_uuid, complete_identity_uuid)
        self.assertNotEqual(viz_uuid, complete_viz_uuid)

        # Assert that the pool was removed
        self.assertFalse(os.path.exists(default_pool_fp))

    def test_mixed_keyed_unkeyed_inputs(self):
        art4_uncached_path = os.path.join(self.tempdir, 'art4.qza')
        self.art4.save(art4_uncached_path)

        self.cache.save(self.art4, 'art4')
        self.cache.save(self.art5, 'art5')

        art4_path = str(self.cache.path) + ':art4'
        art5_path = str(self.cache.path) + ':art5'
        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'dict-of-ints', '--i-ints', f'foo:{art4_path}', '--i-ints',
            art5_path, '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('Keyed values cannot be mixed with unkeyed values.',
                      str(result.exception))

        result = self._run_command(
            'dict-of-ints', '--i-ints', f'foo:{art4_uncached_path}',
            '--i-ints', art5_path, '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('Keyed values cannot be mixed with unkeyed values.',
                      str(result.exception))

        result = self._run_command(
            'dict-of-ints', '--i-ints', f'foo:{art4_path}',
            '--i-ints', art4_uncached_path, '--o-output', output, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('Keyed values cannot be mixed with unkeyed values.',
                      str(result.exception))

    def test_nonexistent_input_key(self):
        art1_path = str(self.cache.path) + ':art1'
        left_path = str(self.cache.path) + ':left'

        result = self._run_command(
            'split-ints', '--i-ints', art1_path, '--o-left', left_path,
            '--o-right', self.non_cache_output, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn("does not contain the key 'art1'",
                      str(result.output))

    def test_output_key_invalid(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        invalid = 'not_valid_identifier$&;'
        out_path = str(self.cache.path) + ':' + invalid
        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', out_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn(f"Key '{invalid}' is not a valid Python identifier",
                      str(result.exception))

    def test_artifact_as_metadata_cache(self):
        self.cache.save(self.mapping, 'mapping')
        mapping_path = str(self.cache.path) + ':mapping'

        result = self.runner.invoke(tools, ['inspect-metadata', mapping_path])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('COLUMN NAME  TYPE', result.output)
        self.assertIn("===========  ===========", result.output)
        self.assertIn("a  categorical", result.output)
        self.assertIn("b  categorical", result.output)
        self.assertIn("IDS:  1", result.output)
        self.assertIn("COLUMNS:  2", result.output)

    def test_artifact_as_metadata_cache_bad_key(self):
        mapping_path = str(self.cache.path) + ':mapping'

        result = self.runner.invoke(tools, ['inspect-metadata', mapping_path])

        self.assertEqual(result.exit_code, 1)
        self.assertIn("does not contain the key 'mapping'", result.output)

    def test_artifact_as_metadata_cache_bad_cache(self):
        result = self.runner.invoke(
            tools, ['inspect-metadata', 'not_a_cache:key'])

        self.assertEqual(result.exit_code, 1)
        self.assertIn('is not a valid cache', result.output)

    def test_output_dir_as_cache(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        out_path = str(self.cache.path) + ':out'

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--output-dir', out_path, '--verbose'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn(
            'Cache keys cannot be used as output dirs.', str(result.exception))

    def test_parallel(self):
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        ints1_path = str(self.cache.path) + ':ints1'

        result = self._run_command(
            'resumable-pipeline', '--i-int-list', ints1_path,
            '--i-int-dict', ints1_path, '--output-dir', output, '--use-cache',
            str(self.cache.path), '--verbose', '--parallel'
        )

        self.assertEqual(result.exit_code, 0)

        list_return = ResultCollection.load(
            os.path.join(output, 'list_return'))
        dict_return = ResultCollection.load(
            os.path.join(output, 'dict_return'))

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        expected = [{
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'}, {
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'
        }]

        self.assertEqual(list_execution_contexts, expected)
        self.assertEqual(dict_execution_contexts, expected)

    def test_config_parallel(self):
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        ints1_path = str(self.cache.path) + ':ints1'

        config_path = get_data_path('mapping_config.toml')

        result = self._run_command(
            'resumable-pipeline', '--i-int-list', ints1_path,
            '--i-int-dict', ints1_path, '--output-dir', output, '--use-cache',
            str(self.cache.path), '--verbose', '--parallel-config', config_path
        )

        self.assertEqual(result.exit_code, 0)

        list_return = ResultCollection.load(
            os.path.join(output, 'list_return'))
        dict_return = ResultCollection.load(
            os.path.join(output, 'dict_return'))

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        list_expected = [{
            'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'}, {
            'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'
        }]
        dict_expected = [{
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'}, {
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'
        }]

        self.assertEqual(list_execution_contexts, list_expected)
        self.assertEqual(dict_execution_contexts, dict_expected)

    def test_both_parallel_flags(self):
        output = os.path.join(self.tempdir, 'output')

        self.cache.save_collection(self.ints1, 'ints1')
        ints1_path = str(self.cache.path) + ':ints1'

        config_path = get_data_path('mapping_config.toml')

        result = self._run_command(
            'resumable-pipeline', '--i-int-list', ints1_path,
            '--i-int-dict', ints1_path, '--output-dir', output, '--use-cache',
            str(self.cache.path), '--verbose', '--parallel',
            '--parallel-config', config_path
        )

        self.assertEqual(result.exit_code, 0)

        list_return = ResultCollection.load(
            os.path.join(output, 'list_return'))
        dict_return = ResultCollection.load(
            os.path.join(output, 'dict_return'))

        list_execution_contexts = self._load_alias_execution_contexts(
            list_return)
        dict_execution_contexts = self._load_alias_execution_contexts(
            dict_return)

        # The explicit config should override the default
        list_expected = [{
            'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'}, {
            'type': 'parsl', 'parsl_type': '_TEST_EXECUTOR_'
        }]
        dict_expected = [{
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'}, {
            'type': 'parsl', 'parsl_type': 'ThreadPoolExecutor'
        }]

        self.assertEqual(list_execution_contexts, list_expected)
        self.assertEqual(dict_execution_contexts, dict_expected)

    def test_parallel_flags_on_non_pipeline(self):
        self.cache.save(self.art1, 'art1')
        self.cache.save(self.art2, 'art2')
        self.cache.save(self.art3, 'art3')

        art1_path = str(self.cache.path) + ':art1'
        art2_path = str(self.cache.path) + ':art2'
        art3_path = str(self.cache.path) + ':art3'

        output = str(self.cache.path) + ':output'

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', output, '--verbose', '--parallel'
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('No such option: --parallel', result.output)

        config_path = get_data_path('mapping_config.toml')

        result = self._run_command(
            'concatenate-ints', '--i-ints1', art1_path, '--i-ints2', art2_path,
            '--i-ints3', art3_path, '--p-int1', '9', '--p-int2', '10',
            '--o-concatenated-ints', output, '--verbose', '--parallel-config',
            config_path
        )

        self.assertEqual(result.exit_code, 1)
        self.assertIn('No such option: --parallel-config', result.output)

    def _load_alias_execution_contexts(self, collection):
        execution_contexts = []

        for result in collection.values():
            alias_uuid = load_action_yaml(
                result._archiver.path)['action']['alias-of']
            execution_contexts.append(load_action_yaml(
                self.cache.data / alias_uuid)
                ['execution']['execution_context'])

        return execution_contexts

    def _load_alias_uuid(self, result):
        return load_action_yaml(result._archiver.path)['action']['alias-of']

    def _load_alias_uuids(self, collection):
        uuids = []

        for artifact in collection.values():
            uuids.append(load_action_yaml(
                artifact._archiver.path)['action']['alias-of'])

        return uuids


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


class DeploymentCache:
    """Cached CLI state for a QIIME deployment.

    In this context, a QIIME deployment is the set of installed Python
    packages, including their exact versions, that register one or more QIIME 2
    plugins. The exact version of q2cli is also included in the deployment.

    The deployment cache stores the current deployment's package names and
    versions in a requirements.txt file under the cache directory. This file is
    used to determine if the cache is outdated. If the cache is determined to
    be outdated, it will be refreshed based on the current deployment state.
    Thus, adding, removing, upgrading, or downgrading a plugin package or q2cli
    itself will trigger a cache refresh.

    Two mechanisms are provided to force a cache refresh. Setting the
    environment variable Q2CLIDEV to any value will cause the cache to be
    refreshed upon instantiation. Calling `.refresh()` will also refresh the
    cache. Forced refreshing of the cache is useful for plugin and/or q2cli
    developers who want their changes to take effect in the CLI without
    changing their package versions.

    Cached CLI state is stored in a state.json file under the cache directory.
    It is not a public file format and it is not versioned. q2cli is included
    as part of the QIIME deployment so that the cached state can always be read
    (or recreated as necessary) by the currently installed version of q2cli.

    This class is intended to be a singleton because it is responsible for
    managing the on-disk cache. Having more than one instance managing the
    cache has the possibility of two instances clobbering the cache (e.g. in a
    multithreaded/multiprocessing situation). Also, having a single instance
    improves performance by only reading and/or refreshing the cache a
    single time during its lifetime. Having two instances could, for example,
    trigger two cache refreshes if Q2CLIDEV is set. To support these use-cases,
    a module-level `CACHE` variable stores a single instance of this class.

    """

    # Public API

    def __init__(self):
        import os

        # Indicates if the cache has been refreshed. For performance purposes,
        # the cache is only refreshed a single time (at maximum) during the
        # object's lifetime. Thus, "hot reloading" isn't supported, but this
        # shouldn't be necessary for the CLI.
        self._refreshed = False

        self._cache_dir = self._get_cache_dir()

        refresh = 'Q2CLIDEV' in os.environ
        self._state = self._get_cached_state(refresh=refresh)

    @property
    def plugins(self):
        """Decoded JSON object representing CLI state on a per-plugin basis."""
        return self._state['plugins']

    def refresh(self):
        """Trigger a forced refresh of the cache.

        If the cache has already been refreshed (either by this method or at
        some point during instantiation), this method is a no-op.

        """
        if not self._refreshed:
            self._state = self._get_cached_state(refresh=True)

    # Private API

    def _get_cache_dir(self):
        import os
        import q2cli.util

        cache_dir = q2cli.util.get_cache_dir()
        os.makedirs(cache_dir, exist_ok=True)
        return cache_dir

    def _get_cached_state(self, refresh):
        import json
        import os.path
        import q2cli.util

        current_requirements = self._get_current_requirements()
        state_path = os.path.join(self._cache_dir, 'state.json')
        # See note on `get_completion_path` for why knowledge of this path
        # exists in `q2cli.util` and not in this class.
        completion_path = q2cli.util.get_completion_path()

        # The cache must be refreshed in the following cases:

        # 1) We have been explicitly told to refresh.
        if refresh:
            self._cache_current_state(current_requirements)
        # 2) The current deployment requirements are different than the cached
        #    requirements.
        elif current_requirements != self._get_cached_requirements():
            self._cache_current_state(current_requirements)
        # 3) The cached state file does not exist.
        elif not os.path.exists(state_path):
            self._cache_current_state(current_requirements)
        # 4) The cached bash completion script does not exist.
        elif not os.path.exists(completion_path):
            self._cache_current_state(current_requirements)

        def decoder(obj):
            if obj.get('__q2type__', None) == 'set':
                return set(obj['value'])
            return obj

        # Now that the cache is up-to-date, read it.
        try:
            with open(state_path, 'r') as fh:
                return json.load(fh, object_hook=decoder)
        except json.JSONDecodeError:
            # 5) The cached state file can't be read as JSON.
            self._cache_current_state(current_requirements)
            with open(state_path, 'r') as fh:
                return json.load(fh, object_hook=decoder)

    # NOTE: The private methods below are all used internally within
    # `_get_cached_state`.

    def _get_current_requirements(self):
        """Includes installed versions of q2cli and QIIME 2 plugins."""
        import os
        import pkg_resources
        import q2cli

        reqs = {
            pkg_resources.Requirement.parse('q2cli == %s' % q2cli.__version__)
        }

        # A distribution (i.e. Python package) can have multiple plugins, where
        # each plugin is its own entry point. A distribution's `Requirement` is
        # hashable, and the `set` is used to exclude duplicates. Thus, we only
        # gather the set of requirements for all installed Python packages
        # containing one or more plugins. It is not necessary to track
        # individual plugin names and versions in order to determine if the
        # cache is outdated.
        #
        # TODO: this code is (more or less) copied from
        # `qiime2.sdk.PluginManager.iter_entry_points`. Importing QIIME is
        # currently slow, and it adds ~600-700ms to any CLI command. This makes
        # the CLI pretty unresponsive, especially when running help/informative
        # commands. Replace with the following lines when
        # https://github.com/qiime2/qiime2/issues/151 is fixed:
        #
        # for ep in qiime2.sdk.PluginManager.iter_entry_points():
        #     reqs.add(ep.dist.as_requirement())
        #
        for entry_point in pkg_resources.iter_entry_points(
                group='qiime2.plugins'):
            if 'QIIMETEST' in os.environ:
                if entry_point.name in ('dummy-plugin', 'other-plugin'):
                    reqs.add(entry_point.dist.as_requirement())
            else:
                if entry_point.name not in ('dummy-plugin', 'other-plugin'):
                    reqs.add(entry_point.dist.as_requirement())

        return reqs

    def _get_cached_requirements(self):
        import os.path
        import pkg_resources

        path = os.path.join(self._cache_dir, 'requirements.txt')

        if not os.path.exists(path):
            # No cached requirements. The empty set will always trigger a cache
            # refresh because the current requirements will, at minimum,
            # contain q2cli.
            return set()
        else:
            with open(path, 'r') as fh:
                contents = fh.read()
            try:
                return set(pkg_resources.parse_requirements(contents))
            except pkg_resources.RequirementParseError:
                # Unreadable cached requirements, trigger a cache refresh.
                return set()

    def _cache_current_state(self, requirements):
        import json
        import os.path
        import click
        import q2cli.core.completion
        import q2cli.util

        click.secho(
            "QIIME is caching your current deployment for improved "
            "performance. This may take a few moments and should only happen "
            "once per deployment.", fg='yellow', err=True)

        cache_dir = self._cache_dir
        state = self._get_current_state()

        path = os.path.join(cache_dir, 'state.json')

        class Q2JSONEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, set):
                    return {
                        '__q2type__': 'set',
                        'value': list(obj),
                    }
                return super().default(obj)

        with open(path, 'w') as fh:
            json.dump(state, fh, cls=Q2JSONEncoder)

        q2cli.core.completion.write_bash_completion_script(
            state['plugins'], q2cli.util.get_completion_path())

        # Write requirements file last because the above steps may raise errors
        # (e.g. a plugin can't be loaded in `_get_current_state`). If any part
        # of the cache writing fails, it needs to be refreshed the next time
        # the cache is accessed. The absence of a requirements file will
        # trigger this cache refresh, avoiding this bug:
        #     https://github.com/qiime2/q2cli/issues/88
        path = os.path.join(cache_dir, 'requirements.txt')
        with open(path, 'w') as fh:
            for req in requirements:
                # `str(Requirement)` is the recommended way to format a
                # `Requirement` that can be read with `Requirement.parse`.
                fh.write(str(req))
                fh.write('\n')

        self._refreshed = True

    def _get_current_state(self):
        """Get current CLI state as an object that is serializable as JSON.

        WARNING: This method is very slow and should only be called when the
        cache needs to be refreshed.

        """
        import q2cli.util

        state = {
            'plugins': {}
        }

        plugin_manager = q2cli.util.get_plugin_manager()
        for name, plugin in plugin_manager.plugins.items():
            state['plugins'][name] = self._get_plugin_state(plugin)

        return state

    def _get_plugin_state(self, plugin):
        import q2cli.core.state

        state = q2cli.core.state.get_plugin_state(plugin)
        for id, action in plugin.actions.items():
            state['actions'][id]['epilog'] = self._get_action_epilog(action)

        return state

    def _get_action_epilog(self, action):
        import q2cli.core.usage

        lines = []
        for name, example in action.examples.items():
            use = q2cli.core.usage.CLIUsage()

            use.comment('### example: %s\n' % (name.replace('_', ' '),))
            example(use)
            use.recorder.append('')

            lines += use.recorder

        return lines


# Singleton. Import and use this instance as necessary.
CACHE = DeploymentCache()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

import q2cli.commands


ROOT_COMMAND_HELP = """\
QIIME 2 command-line interface (q2cli)
--------------------------------------

To get help with QIIME 2, visit https://qiime2.org.

To enable tab completion in Bash, run the following command or add it to your \
.bashrc/.bash_profile:

    source tab-qiime

To enable tab completion in ZSH, run the following commands or add them to \
your .zshrc:

\b
    autoload -Uz compinit && compinit
    autoload bashcompinit && bashcompinit
    source tab-qiime

"""


# Entry point for CLI
@click.command(cls=q2cli.commands.RootCommand, invoke_without_command=True,
               no_args_is_help=True, help=ROOT_COMMAND_HELP)
@click.version_option(prog_name='q2cli',
                      message='%(prog)s version %(version)s\nRun `qiime info` '
                              'for more version details.')
def qiime():
    pass


if __name__ == '__main__':
    qiime()
